#include "../df_lua_api.h"
#include <algorithm>

#include "../df_lua_connection.h"
#include "../df_lua_files.h"
#include "../df_lua_setup_records.h"
#include "../df_lua_active_mode.h"

int l_readFirmwareVersion( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );
    // Device
    int deviceAddress = static_cast<int>( luaL_checkinteger( L, 2 ) );
    luaL_argcheck( L, deviceAddress >= 1 && deviceAddress <= 254, 2, "out of range" );

    std::string version( 64, '\0' );
    int res = DFCGetVersionFirmware( lc->id, deviceAddress, const_cast<char *>( version.data() ), &lc->errorNumber );
    if ( res == 0 )
    {
        lua_pushnil( L );
        res = 1;
        push_errorNumber( lc, res );
    }
    else
    {
        lua_pushlstring( L, version.data(), version.find_first_of( '\0' ) );
        res = 1;
    }

    return res;
}

int l_getLibraryVersion( lua_State *L )
{
    std::string version( 64, '\0' );

    if ( DFCVersionLibrary( const_cast<char *>( version.data() ), version.size() ) == 0 )
        lua_pushnil( L );
    else
        lua_pushlstring( L, version.data(), version.find_first_of( '\0' ) );

    return 1;
}

int l_writeDisplayMessage( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );
    // Device
    int deviceAddress = static_cast<int>( luaL_checkinteger( L, 2 ) );
    luaL_argcheck( L, deviceAddress >= 1 && deviceAddress <= 254, 2, "out of range" );
    // visibleTime
    int visibleTime = static_cast<int>( luaL_checkinteger( L, 3 ) );
    luaL_argcheck( L, visibleTime >= 0 && visibleTime <= 255, 3, "out of range" );
    // audibleSignal
    int audibleSignal = static_cast<int>( luaL_checkinteger( L, 4 ) );
    luaL_argcheck( L, audibleSignal >= 0 && audibleSignal <= 255, 4, "out of range" );
    // text
    const char *text = luaL_checkstring( L, 5 );
    // doAction
    int doAction = static_cast<int>( luaL_optinteger( L, 4, 0 ) );

    int res = 1;

    lua_pushboolean( L, DFCComSendMessage( lc->id, deviceAddress, visibleTime, doAction, audibleSignal, const_cast<char *>( text ), std::strlen( text ) ) );

    if ( lua_toboolean( L, -1 ) == false )
    {
        lc->errorNumber = IV_ERROR_UNREACHABLE;
        push_errorNumber( lc, res );
    }

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCComSendInfotext( int connectionId, int deviceAddress, char *text, int textLength )
int l_writeDisplayMemo( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );
    // Device
    int deviceAddress = static_cast<int>( luaL_checkinteger( L, 2 ) );
    luaL_argcheck( L, deviceAddress >= 1 && deviceAddress <= 254, 2, "out of range" );
    // text
    const char *text = luaL_checkstring( L, 3 );

    int res = 1;

    lua_pushboolean( L, DFCComSendInfotext( lc->id, deviceAddress, const_cast<char *>( text ), std::strlen( text ) ) );

    if ( lua_toboolean( L, -1 ) == false )
    {
        lc->errorNumber = IV_ERROR_UNREACHABLE;
        push_errorNumber( lc, res );
    }

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCComSetTime( int connectionId, int deviceAddress, unsigned char *dateTimeBuffer )
//DF_API_EXTERN int DF_API_CALL DFCComGetTime( int connectionId, int deviceAddress, unsigned char *dateTimeBuffer )
//DF_API_EXTERN int DF_API_CALL DFCGetSeriennummer( int connectionId, int deviceAddress, int *errorNumber, int *value )

//DF_API_EXTERN int DF_API_CALL DFCSetGlobVar( int connectionId, int deviceAddress, unsigned char *name, int nameType, unsigned char *value, int *errorNumber )
int l_writeFirmwareVariable( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );
    // Device
    int deviceAddress = static_cast<int>( luaL_checkinteger( L, 2 ) );
    luaL_argcheck( L, deviceAddress >= 1 && deviceAddress <= 254, 2, "out of range" );
    // name
    std::string name( luaL_checkstring( L, 3 ) );
    // value
    const char *value = luaL_checkstring( L, 4 );

    int res = 1;
    int nameType = 2;
    std::string::size_type pos = name.find_first_of( '.' );
    luaL_argcheck( L, pos != std::string::npos, 3, "no '.' found in name" );

    std::string major( name.substr( 0, pos ) );
    std::transform( major.begin(), major.end(), major.begin(), ::tolower );
    if ( major == "setup" )
    {
        nameType = 1;
        name = name.substr( pos + 1 );
        if ( std::all_of( name.begin(), name.end(), ::isdigit ) )
        {
            nameType = 0;
            name[0] = std::atoi( name.data() );
            name.resize( 1 );
        }
    }

    lua_pushboolean( L, DFCSetGlobVar( lc->id, deviceAddress, reinterpret_cast<unsigned char*>( const_cast<char *>( name.data() ) ), nameType, reinterpret_cast<unsigned char*>( const_cast<char *>( value ) ), &lc->errorNumber ) );

    if ( lua_toboolean( L, -1 ) == false )
        push_errorNumber( lc, res );

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCGetGlobVar( int connectionId, int deviceAddress, unsigned char *name, int nameType, unsigned char *value, int valueLength, int *errorNumber )
int l_readFirmwareVariable( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );
    // Device
    int deviceAddress = static_cast<int>( luaL_checkinteger( L, 2 ) );
    luaL_argcheck( L, deviceAddress >= 1 && deviceAddress <= 254, 2, "out of range" );
    // name
    std::string name( luaL_checkstring( L, 3 ) );
    // value
    std::string value( 1024, '\0' );
    
    int nameType = 2;
    std::string::size_type pos = name.find_first_of( '.' );
    luaL_argcheck( L, pos != std::string::npos, 3, "no '.' found in name" );

    std::string major( name.substr( 0, pos ) );
    std::transform( major.begin(), major.end(), major.begin(), ::tolower );
    if ( major == "setup" )
    {
        nameType = 1;
        name = name.substr( pos + 1 );
        if ( std::all_of( name.begin(), name.end(), ::isdigit ) )
        {
            nameType = 0;
            name[0] = std::atoi( name.data() );
            name.resize( 1 );
        }
    }

    int res = DFCGetGlobVar( lc->id, deviceAddress, reinterpret_cast<unsigned char*>( const_cast<char *>( name.data() ) ), nameType, reinterpret_cast<unsigned char*>( const_cast<char *>( value.data() ) ), value.size(), &lc->errorNumber );
    if ( res == 0 )
    {
        lua_pushnil( L );
        res = 1;
        push_errorNumber( lc, res );
    }
    else
    {
        lua_pushlstring( L, value.data(), value.find_first_of( '\0' ) );
        res = 1;
    }

    return res;
}

int l_getChannelError( lua_State *L )
{
    // Channel-Data
    LuaChannel *lc = checkchannel( L );

    int res = 0;
    push_errorNumber( lc, res );

    return res;
}

//DF_API_EXTERN int DF_API_CALL DFCCloseRelay( int connectionId, int deviceAddress, int number, int timeout, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCGetRelayState( int connectionId, int deviceAddress, int number, int *state, int *timeout, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCOpenRelay( int connectionId, int deviceAddress, int number, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCGetInfo( int connectionId, int deviceAddress, char *name, char *params, char *text, int *textLength, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCSetFontType( int connectionId, int deviceAddress, int type, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCPressVirtualKey( int connectionId, int deviceAddress, int keyCode, int flags, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCGetPasswordKey( int connectionId, int deviceAddress, char *value, int *valueLength, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCSetPassword( int connectionId, char *value, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCReset( int connectionId, int deviceAddress, int mode, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCSetCommunicationPassword( int connectionId, unsigned char *value, int valueLength, int sessionTimeout, int *errorNumber )
//DF_API_EXTERN int DF_API_CALL DFCRecordVolume( int connectionId, int deviceAddress, int *recordCount, int *memoryUsage, int *memorySize, int *errorNumber )
//
//

int l_getChannel( lua_State *L )
{
    int id = static_cast<int>( luaL_checkinteger( L, 1 ) );
    luaL_argcheck( L, id >= 1 && id < (MAX_CHANNELID - 1), 1, "out of range" );

    LuaChannel *lc = reinterpret_cast<LuaChannel *>( lua_newuserdata( L, sizeof(LuaChannel) ) );
    lc->id = id;

    luaL_getmetatable( L, "datafox.channel" );
    lua_setmetatable( L, -2 );

    return 1;
}

int l_tostring( lua_State *L )
{
    LuaChannel *lc = checkchannel( L );
    lua_pushfstring( L, "channel(%d)", lc->id );

    return 1;
}

const struct luaL_Reg dfcomdll_f[] = {
    { "getChannel", l_getChannel },
    { "getLibraryVersion", l_getLibraryVersion },
    // Active-Mode
    { "startActiveMode", l_startActiveMode },
    { "stopActiveMode", l_stopActiveMode },
    { "getActiveModeFirstId", l_getActiveModeFirstId },
    { "getActiveModeNextId", l_getActiveModeNextId },
    { "getActiveModeInfo", l_getActiveModeInfo },
    { "setActiveModeNotify", l_setActiveModeNotify },
    { "getActiveModeNotifyAvailable", l_getActiveModeNotifyAvailable },
    { "bindActiveModeChannel", l_bindActiveModeChannel },
    // Files
    { "editFile", l_editFile },
    { NULL, NULL }
};

const struct luaL_Reg dfcomdll_m[] = {
    // Connection
    { "open", l_open },
    { "close", l_close },
    { "isOpen", l_isOpen },
    { "write", l_write },
    { "read", l_read },
    // Files
    { "writeFile", l_writeFile },
    { "readFile", l_readFile },
    // Records
    { "readRecordTables", l_readRecordTables },
    { "getRecordTableCount", l_getRecordTableCount },
    { "getRecordTable", l_getRecordTable },
    { "getRecordField", l_getRecordField },
    { "readRecord", l_readRecord },
    { "quitRecord", l_quitRecord },
    { "resetRecords", l_resetRecords },
    { "restoreRecords", l_restoreRecords },
    // Api
    { "readFirmwareVersion", l_readFirmwareVersion },
    { "writeDisplayMessage", l_writeDisplayMessage },
    { "writeDisplayMemo", l_writeDisplayMemo },
    { "writeFirmwareVariable", l_writeFirmwareVariable },
    { "readFirmwareVariable", l_readFirmwareVariable },
    { "getChannelError", l_getChannelError },
    // Utils
    { "__tostring", l_tostring },
    { NULL, NULL }
};

int luaopen_dfcomdll( lua_State *L )
{
    luaL_newmetatable( L, "datafox.channel" );
    lua_pushvalue( L, -1 );
    lua_setfield( L, -2, "__index" );

    luaL_setfuncs( L, dfcomdll_m, 0 );

    luaL_newlib( L, dfcomdll_f );

    return 1;
}
