#include "fileintelhex.h"
#include "../Common/crc.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <memory.h>

FileIntelHex::EnumErrorState FileIntelHex::_errorState = FileIntelHex::None;
std::string FileIntelHex::_errorText = "";
std::string FileIntelHex::_errorDetail = "";

FileIntelHex::FileIntelHex() :
    _maxDataSize( 20 ),
    _fillByte( 0xFF ),
    _blockSize( 0x10000 )
{
    // Fr die CRC-Berechnung Initalisieren.
    InitCRC32Table();
    // Initialisieren
    reset();
    // Fehlerobjekt zuruecksetzen
    resetError();

    // Der Standardmig zu verarbeitende Datensatztyp
    setRecordType( TypeDAT );
}

FileIntelHex::~FileIntelHex()
{
}

void FileIntelHex::resetError()
{
	// Member initialisieren
	_errorState = None;
	_errorText.clear();
	_errorDetail.clear();
}

void FileIntelHex::reset()
{
    // Angabe der Datei
    setFilePath( "" );

    // Register und Puffer
    _startSegmentAddress.ipRegister = 0;
    _startSegmentAddress.csRegister = 0;
    _startSegmentAddress.exists = false;

    _startLinearAddress.eipRegister = 0;
    _startLinearAddress.exists = false;

    _baseAddress = 0;
    _addressMode = Unknown;
    _bufferFillBytes.clear();
    _buffer.clear();
}

bool FileIntelHex::load( const std::string &filePath )
{
    bool res = true;

    // Lesen der Hexdatei in die internen Speicher.
    std::ifstream fileIn( filePath );
    std::string lineBuffer;
    std::vector<bool> tempFillBytes;
    std::vector<unsigned char> tempBuffer;

    bool initAddresses = false;
    unsigned int numLine = 0;
    unsigned int extendedAddress = 0;
    unsigned int maxAddress = 0;
    unsigned short recordAddress;
    unsigned char byteCount;
    unsigned char recType;

    // Umgebung zuruecksetzen.
    reset();

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( !fileIn.is_open() )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( FileOpen );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Zeilenweise lesen, bis eof oder Fehler.
    while( !fileIn.eof() && (errorState() == None) )
    {
        // Zeile lesen.
        if ( getline( fileIn, lineBuffer ) )
        {
            // Anzahl der Zeilen.
            numLine++;

            recType = GetType( lineBuffer.data() );
            // Je nach Datensatztyp verarbeiten.
            switch ( recType )
            {
            case TypeDAT:                // Data record
            case TypeDisplayFont:        // Datei der Anzeigedaten
            case TypeDisplayGrafic:
            case TypeDeviceHeader:       // Datei der Geraetedaten
            case TypeTextData:           // Datei der Textdaten
            case TypeTouchkeyData:       // Datei der Touchkeydaten
            case TypeDisplayColorData:   // Datei der Displayfarben
            case TypeWlanConfigData:     // Datei der WlanConfig
            case TypeHardwareConfigData: // Datei der HardwareConfig
            {
                // Nur diesen Datensatztypen verarbeiten.
                if ( recordType() != recType )
                    continue;

                recordAddress = GetAddress( lineBuffer.data() );
                byteCount = GetCount( lineBuffer.data() );
                tempBuffer.resize( byteCount );

                // Erkennung ob bereits die Basisaddresse initalisiert ist.
                if ( !initAddresses )
                {
                    // Basisadresse fr die Daten halten.
                    _baseAddress = extendedAddress + recordAddress;
                    initAddresses = true;
                }

                // Wird ggf. vor der bisherigen Adresse eingefgt.
                if ( _baseAddress > (extendedAddress + recordAddress) )
                {
                    // Diese Daten mssen vor den bisherigen eingefgt werden.

                    // Die bisherigen Daten mssen um die byteCount nach hinten verschoben werden.
                    _bufferFillBytes.resize( tempBuffer.size() );
                    std::fill_n( tempFillBytes.begin(), tempFillBytes.size(), false );
                    _bufferFillBytes.insert( _bufferFillBytes.begin(), tempFillBytes.begin(), tempFillBytes.end() );

                    std::fill_n( tempBuffer.begin(), tempBuffer.size(), fillByte() );
                    _buffer.insert( _buffer.begin(), tempBuffer.begin(), tempBuffer.end() );

                    // Basisadresse neu setzen.
                    _baseAddress = extendedAddress + recordAddress;
                }

                // Eventuell nchsten Speicherblock reservieren.
                if ( ((extendedAddress + recordAddress + tempBuffer.size()) - _baseAddress) > _buffer.size() )
                {
                    _bufferFillBytes.resize( _bufferFillBytes.size() + blockSize(), false );
                    _buffer.resize( _buffer.size() + blockSize(), fillByte() );
                }

                GetData( lineBuffer.data(), tempBuffer.data() );

                std::vector<bool>::iterator iterFillBytesDest = _bufferFillBytes.begin() + ((extendedAddress + recordAddress) - _baseAddress);
                std::vector<unsigned char>::iterator iterDest = _buffer.begin() + ((extendedAddress + recordAddress) - _baseAddress);
                std::vector<unsigned char>::iterator iterSource = tempBuffer.begin();
                for ( ; iterSource != tempBuffer.end(); ++iterDest, ++iterFillBytesDest, ++iterSource )
                {
                    if ( (*iterFillBytesDest) )
                    {
                        // Diese Addresse wurde bereits beschrieben.
                        setErrorState( FileFormat );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        return false;
                    }

                    (*iterDest) = (*iterSource);
                    (*iterFillBytesDest) = true; // Merken, dass dieses Byte eingelesen wurde.
                }

                // Maximal bearbeitete Adresse halten.
                if ( (extendedAddress + recordAddress + tempBuffer.size()) > maxAddress )
                    maxAddress = extendedAddress + recordAddress + tempBuffer.size();

                break;}

            case TypeEOF: // End-of-file record
                break;

            case TypeESA:{ // Extended Segment Address Record
                if ( addressMode() == Unknown )
                {
                    setAddressMode( Segment );
                }
                else if ( addressMode() != Segment )
                {
                    // Es geht nur entweder Segmentiert oder Linear.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return false;
                }
                // Dieses ist nun die aktuelle Basisadresse fr weitere Daten (TypeDAT)
                extendedAddress = GetData16( lineBuffer.data(), 0 ) << 4;
                break;}

            case TypeSSA: // Start Segment Address Record
                if ( addressMode() == Unknown )
                {
                    setAddressMode( Segment );
                }
                else if ( addressMode() != Segment )
                {
                    // Es geht nur entweder Segmentiert oder Linear.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return false;
                }
                if ( _startLinearAddress.exists )
                {
                    // Start Linear Address ist bereits definiert, kann nur einmal geschehen.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return false;
                }
                _startSegmentAddress.csRegister = GetData16( lineBuffer.data(), 0 );
                _startSegmentAddress.ipRegister = GetData16( lineBuffer.data(), 2 );
                _startSegmentAddress.exists = true;
                break;

            case TypeELA: { // Extended Linear Address Record
                if ( addressMode() == Unknown )
                {
                    setAddressMode( Linear );
                }
                else if ( addressMode() != Linear )
                {
                    // Es geht nur entweder Segmentiert oder Linear.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return false;
                }
                // Dieses ist nun die aktuelle Basisadresse fr weitere Daten (TypeDAT)
                extendedAddress = GetData16( lineBuffer.data(), 0 ) << 16;
                break;}

            case TypeSLA: // Start Linear Address Record
                if ( addressMode() == Unknown )
                {
                    setAddressMode( Linear );
                }
                else if ( addressMode() != Linear )
                {
                    // Es geht nur entweder Segmentiert oder Linear.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    return false;
                }

                if ( _startSegmentAddress.exists )
                {
                    // Start Segment Address ist bereits definiert, kann nur einmal geschehen.
                    setErrorState( FileFormat );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                }
                _startLinearAddress.eipRegister = GetData16( lineBuffer.data(), 0 ) << 16;
                _startLinearAddress.eipRegister += GetData16( lineBuffer.data(), 2 );
                _startLinearAddress.exists = true;
                break;
            }
        }
    }

    // Datenpuffer an die letze Adresse anpassen.
    _bufferFillBytes.resize( maxAddress - _baseAddress );
    _buffer.resize( maxAddress - _baseAddress );

    // Geladende Dateiangabe halten.
    setFilePath( filePath );

    return true;
}

bool FileIntelHex::save( const std::string &filePath, bool append )
{
    // Schreiben der internen Speicher in die Hex-Datei.
    std::ofstream fileOut( filePath, append ? std::ios::app : std::ios::out );
    std::string lineBuffer( INTELHEX_LENGTH_MAX, '\0' );

    unsigned int numLine = 0;
    unsigned int currentAddress;
    unsigned int extendedAddress = ~0;
    unsigned short recordAddress;

    std::vector<unsigned char>::iterator iterContent;
    std::vector<unsigned char> recordData;
    std::vector<unsigned char> addressData;

    // Fehlerobjekt zuruecksetzen
    resetError();

    if ( !fileOut.is_open() )
    {
        // Oeffnen der Datei gescheitert
        setErrorState( FileOpen );
        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Iterator auf den Datenbegin.
    iterContent = _buffer.begin();
    // Aktuelle Adresse zuweisen.
    currentAddress = _baseAddress;

    // Schreiben der Anwendungsfirmware
    for ( ; ; ++iterContent, ++currentAddress )
    {
        // Datensatz bei folgenden Bedingungen schreiben.
        // - Puffer der Daten hat sein Ende erreicht, Angesammelte Daten schreiben.
        if ( iterContent == _buffer.end() )
        {
            if ( recordData.size() )
            {			
                // Angefangenen Datenrecord schreiben
                if ( !MakeTypedDataLine( (currentAddress & 0xFFFF) - recordData.size(), recordType(), recordData.data(), recordData.size(), const_cast<char*>( lineBuffer.data() ) ) )
                {
                    // Fehler zuweisen
                    setErrorState( RecordMake );
                    setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                    break;
                }

                //fileOut << lineBuffer.c_str() << std::endl;
                fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;

                numLine++;
            }

            // Abbruchbedingung
            break;
        }

        // Datensatz bei folgenden Bedingungen schreiben.
        // - Wechsel der Erweiterten Addresse fr aktuelles Datenbyte, Angesammelte Daten schreiben.
        if (  recordData.size() && (currentAddress & 0xFFFF0000) != extendedAddress )
        {
            // Angefangenen Datenrecord schreiben
            if ( !MakeTypedDataLine( (currentAddress & 0xFFFF) - recordData.size(), recordType(), recordData.data(), recordData.size(), const_cast<char*>( lineBuffer.data() ) ) )
            {
                // Fehler zuweisen
                setErrorState( RecordMake );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                break;
            }

            //fileOut << lineBuffer.c_str() << std::endl;
            fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
            numLine++;

            // Datensatzpuffer leeren
            recordData.clear();
        }

        // Wechsel der Erweiterten Addresse?
        if ( (currentAddress & 0xFFFF0000) != extendedAddress )
        {	
            // Nun aktuelle erweiterte Adresse zuweisen.
            extendedAddress = currentAddress & 0xFFFF0000;

            // Erweiterte Adresse je nach Modus schreiben.
            switch ( addressMode() )
            {
            case Segment:
                {
                    recordAddress = (extendedAddress >> 4) & 0xFFFF;
                    addressData.clear();
                    addressData.push_back( (recordAddress >> 8) & 0xFF );
                    addressData.push_back( recordAddress & 0xFF );

                    // Extended Segment Address Record
                    if ( !MakeTypedDataLine( 0, TypeESA, addressData.data(), addressData.size(), const_cast<char*>( lineBuffer.data() ) ) )
                    {  
                        // Fehler zuweisen
                        setErrorState( RecordMake );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        return false;
                    }

                    //fileOut << lineBuffer.c_str() << std::endl;
                    fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
                    numLine++;
                    break;
                }

            case Linear:
                {
                    recordAddress = (extendedAddress >> 16) & 0xFFFF;
                    addressData.clear();
                    addressData.push_back( (recordAddress >> 8) & 0xFF );
                    addressData.push_back( recordAddress & 0xFF );

                    // Extended Segment Address Record
                    if ( !MakeTypedDataLine( 0, TypeELA, addressData.data(), addressData.size(), const_cast<char*>( lineBuffer.data() ) ) )
                    {  
                        // Fehler zuweisen
                        setErrorState( RecordMake );
                        setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                        return false;
                    }

                    //fileOut << lineBuffer.c_str() << std::endl;
                    fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
                    numLine++;
                    break;
                }
            }
        }		

        // Aktuelles Datenbyte anfgen.
        recordData.push_back( (*iterContent) );

        // Datensatz bei folgenden Bedingungen schreiben.
        // - Maximale Anzahl Datenbyte pro Datensatz erreicht.
        if ( recordData.size() == maxDataSize() )
        {
            // Angefangenen Datenrecord schreiben
            if ( !MakeTypedDataLine( (currentAddress & 0xFFFF) - recordData.size() + 1, recordType(), recordData.data(), recordData.size(), const_cast<char*>( lineBuffer.data() ) ) )
            {
                // Fehler zuweisen
                setErrorState( RecordMake );
                setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                break;
            }

            //fileOut << lineBuffer.c_str() << std::endl;
            fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
            numLine++;

            // Datensatzpuffer leeren
            recordData.clear();
        }
    }

    if ( _startSegmentAddress.exists )
    {
        // Start Segment Address Record schreiben.

        addressData.clear();
        addressData.push_back( (_startSegmentAddress.csRegister >> 8) & 0xFF );
        addressData.push_back( _startSegmentAddress.csRegister & 0xFF );
        addressData.push_back( (_startSegmentAddress.ipRegister >> 8) & 0xFF );
        addressData.push_back( _startSegmentAddress.ipRegister & 0xFF );
        // Start Segment Address Record
        if ( !MakeTypedDataLine( 0, TypeSSA, addressData.data(), addressData.size(), const_cast<char*>( lineBuffer.data() ) ) )
        {  
            // Fehler zuweisen
            setErrorState( RecordMake );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }

        //fileOut << lineBuffer.c_str() << std::endl;
        fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
        numLine++;
    }
    else if ( _startLinearAddress.exists )
    {
        // Start Linear Address Record schreiben.

        addressData.clear();
        addressData.push_back( (_startLinearAddress.eipRegister >> 24) & 0xFF );
        addressData.push_back( (_startLinearAddress.eipRegister >> 16) & 0xFF );
        addressData.push_back( (_startLinearAddress.eipRegister >> 8) & 0xFF );
        addressData.push_back( _startLinearAddress.eipRegister & 0xFF );
        // Start Linear Address Record
        if ( !MakeTypedDataLine( 0, TypeSLA, addressData.data(), addressData.size(), const_cast<char*>( lineBuffer.data() ) ) )
        {  
            // Fehler zuweisen
            setErrorState( RecordMake );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }

        //fileOut << lineBuffer.c_str() << std::endl;
        fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
        numLine++;

    }

    if ( !append )
    {
        if ( !MakeEndLine( const_cast<char*>( lineBuffer.data() ) ) )
        {
            // Fehler zuweisen
            setErrorState( RecordEndMake );
            setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return false;
        }

        //fileOut << lineBuffer.c_str() << std::endl;
        fileOut.write( lineBuffer.data(), lineBuffer.find( '\0' ) ) << std::endl;
        numLine++;
    }

    // Geschriebene Dateiangabe halten.
    setFilePath( filePath );

    return true;
}

unsigned int FileIntelHex::checksum32bit()
{
    unsigned int crc = -1;

    // Workaround fr einen Bug bei den aktuellen Geraetedateien.
    if ( firstAddress() > 0 && _buffer.size() > firstAddress() )
    {
        // Hierbei wurde die Startaddresse nicht bei der CRC-Berechnung beachtet.

        std::vector<unsigned char> tmp( firstAddress(), 0xFF );
        std::copy_n( _buffer.begin(), _buffer.size() - firstAddress(), std::back_inserter( tmp ) );

        // CRC berechnen.
        return GetCRC32( tmp.data(), tmp.size(), crc );
    }

    // CRC berechnen.
    if ( _buffer.size() )
        crc = GetCRC32( _buffer.data(), _buffer.size(), crc );

    return crc;
}

// ########################################################
// HELPER

unsigned short FileIntelHex::HexToUINT16( const char *data )
{
	int i;
	unsigned short val = 0;

	for ( i = 0; i < 4; ++i )
	{
		val = (val << 4);

		if ( data[i] >= '0' && data[i] <= '9' )
			val += data[i] - '0';
		else if ( data[i] >= 'a' && data[i] <='f' )
			val += data[i] - 'a' + 10;
		else
			val += data[i] - 'A' + 10;
	}

	return val;
}

void FileIntelHex::UINT16ToHex( unsigned short val, char *data )
{
	int i;

	for ( i = 3; i >= 0; --i )
	{
		data[i] = val & 0xF;

		if ( data[i] > 9 )
			data[i] += 'A' - 10; 
		else
			data[i] += '0';

		val >>= 4;
	}	
}

unsigned char FileIntelHex::HexToUINT8( const char *data )
{
	int i;
	unsigned char val = 0;

	for ( i = 0; i < 2; ++i )
	{
		val = (val << 4);

		if ( data[i] >= '0' && data[i] <= '9' )
			val += data[i] - '0';
		else if ( data[i] >= 'a' && data[i] <='f' )
			val += data[i] - 'a' + 10;
		else
			val += data[i] - 'A' + 10;
	}

	return val;
}

void FileIntelHex::UINT8ToHex( unsigned char val, char *data )
{
	data[0] = (val >> 4) & 0xF;

	if ( data[0] > 9 )
		data[0] += 'A' - 10; 
	else
		data[0] += '0';

	data[1] = val & 0xF;

	if ( data[1] > 9 )
		data[1] += 'A' - 10;
	else
		data[1] += '0';
}

int FileIntelHex::MakeCrc( char *data )
{
	int length = 0, lastByte, check = 0;

	if ( data[length] != INTELHEX_START_SIGN )
		return 0;

	length++;

	// Laenge der Datenzeile ermitteln
	while( isxdigit( (int)data[length] ) )
	{
		length++;

		if ( !(length % 2) )
		{
			lastByte = HexToUINT8( &data[length - 1] );
			check += lastByte;
		}
	}

	if ( length < 9 ) // Mindestens 9 Zeichen: ":00000001"
		return 0;

	// Pruefsumme eintragen
	UINT8ToHex( (unsigned char)(~(((char*)&check)[0] - 1) & 0xFF), &data[length] );

	return 1;
}

int FileIntelHex::CheckCrc( const char *data )
{
	int length = 0, lastByte = 0, check = 0;

	if ( data[length] != INTELHEX_START_SIGN )
		return 0;

	length++;

	// Laenge der Datenzeile ermitteln
	while( isxdigit( (int)data[length]) )
	{
		length++;

		if ( !(length % 2) )
		{
			lastByte = HexToUINT8( &data[length - 1] );
			check += lastByte;
		}
	}

	if ( length < 11 ) // Mindestens 11 Zeichen: ":00000001FF"
		return 0;

	// Pruefsumme aus Gesamtsumme entfernen
	check -= lastByte;
	// Zweierkomplement
	check = ~(((char*)&check)[0] - 1) & 0xFF;

	return (check == lastByte);
}

int FileIntelHex::MakeTypedDataLine( unsigned short address, unsigned char type, unsigned char *buf, unsigned char length, char *data )
{
	data[INTELHEX_START] = INTELHEX_START_SIGN;      // start of record
	UINT8ToHex( length, &data[INTELHEX_COUNT] );     // byte count
	UINT16ToHex( address, &data[INTELHEX_ADDRESS] ); // load address
	UINT8ToHex( type, &data[INTELHEX_TYPE] );        // record type

	// insert data
	unsigned char i;
	for ( i = 0; i < length; ++i )
		UINT8ToHex( buf[i], &data[INTELHEX_DATA + (i * 2)] );

	// Endezeichen fuer CRC - Laengenermittlung
	data[INTELHEX_DATA + (i * 2)] = 0;

	// checksum suffix
	if ( MakeCrc( data ) )
	{
		i++;

		// Nullterminie
		data[INTELHEX_DATA + (i * 2)] = 0;
		return 1;
	}

	return 0;
}

int FileIntelHex::MakeDataLine( unsigned short address, unsigned char *buf, unsigned char length, char *data )
{
	return MakeTypedDataLine( address, TypeDAT, buf, length, data );
}

int FileIntelHex::MakeEndLine( char *data )
{
	data[INTELHEX_START] = INTELHEX_START_SIGN;            // start of record
	UINT8ToHex( 0, &data[INTELHEX_COUNT] );                // byte count
	UINT16ToHex( 0, &data[INTELHEX_ADDRESS] );             // load address
	UINT8ToHex( TypeEOF, &data[INTELHEX_TYPE] ); // record type

	// Endezeichen fuer CRC - Laengenermittlung
	data[INTELHEX_DATA] = 0;

	// checksum suffix
	if ( MakeCrc( data ) )
	{
		// Nullterminie
		data[INTELHEX_DATA + 2] = 0;
		return 1;
	}

	return 0;
}

// ########################################################
// Sucht den Anfang und fuehrt grundlegende Pruefungen durch.
char *FileIntelHex::PrepareDataLine( char *data, int length )
{
	int pos;
	// Datensatzanfang suchen
	for ( pos = 0; data[pos] != '\0' && pos < length; ++pos )
	{
		// Startzeichen?
		if ( data[pos] == INTELHEX_START_SIGN )
		{
			// Liegt der gesamte Datensatz im Puffer vor?
			if ( (INTELHEX_LENGTH_FRAME + GetCount( &data[pos] ) * 2) > (length - pos) )
			{
				setErrorState( RecordBufferOverflow );
				setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
				return NULL;
			}

			// Checksumme pruefen
			if ( !CheckCrc( &data[pos] ) )
			{
				setErrorState( RecordCrc );
				setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
				return NULL;
			}

			return &data[pos];
		}
		else if ( data[pos] != '\n' && data[pos] != '\t' && data[pos] != ' ' )
		{
			// Zwischen den datenrecords duerfen nur Whitespaces liegen
			setErrorState( RecordFormat );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			return NULL;
		}
	}

	setErrorState( RecordEmpty );
	setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
	return NULL;
}

// ########################################################
// Laden und Pruefen der Datei.
int FileIntelHex::CheckContent( const char *fileName )
{
	size_t pos = 0;
	unsigned int numRecord = 0;
	char buf[530], *cur = NULL, *curTmp;

	FILE *fp = fopen( fileName, "r" );

	if ( fp == NULL )
	{
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
		return 0;
	}

	// Leerer Datenpuffer als Vorgabe
	setErrorState( RecordEmpty );
	setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );

	do
	{
		if ( errorState() == RecordBufferOverflow )
		{
			pos = sizeof(buf) - (cur - buf);
			// Datensatzrest an Pufferanfang kopieren
			memcpy( buf, cur, pos );
			// Datenpuffer leeren und fuellen
			memset( &buf[pos], 0, sizeof(buf) - pos );
			fread( &buf[pos], sizeof(buf) - pos, 1, fp );
			// Datenzeiger setzen
			cur = buf;
		}
		else if ( errorState() == RecordEmpty )
		{
			// Datenpuffer fuellen
			memset( buf, 0, sizeof(buf) );
			fread( buf, sizeof(buf), 1, fp );
			// Datenzeiger setzen
			cur = buf;
		}
		else
		{
			// Es ist ein Verarbeitungsfehler eingetreten.
			break;
		}

		if( ferror( fp ) )
		{
			// Lag ein Dateifehler vor?
			setErrorState( FileRead );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
			break;
		}

		// Alle gueltigen Datensaetze im Puffer pruefen
		while( (curTmp = PrepareDataLine( cur, int(sizeof(buf) - (cur - buf) ) )) != NULL )
		{
			// Records zaehlen
			numRecord++;
			// Recordstart uebernehmen
			cur = curTmp;
			// Hinter aktuellen Datenrecord setzen
			cur += INTELHEX_LENGTH_FRAME + GetCount( cur ) * 2;
		}

	}
	while( !feof( fp ) );

	// Wurde die Datei komplett gelesen und der Puffer ist leer, dann 
	// ist die Datei gueltig, wenn mindestens ein Datensatz vorhanden war.
	if ( feof( fp ) && errorState() == RecordEmpty && numRecord )
	{
		resetError();
	}
	else if ( errorState() == RecordEmpty && numRecord == 0 )
	{
		// Es liegt kein Datensatz vor, die Datei ist somit leer.
		setErrorState( FileEmpty );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
	}

	// Datei schliessen
	fclose( fp );

	return (errorState() == None);
}

// ########################################################
// Pruefen der Datei mit Inhalt.
int FileIntelHex::CheckFile( const char *fileName )
{
	unsigned int pos = 0, numRecord = 0;
	char buf[530];
	unsigned char fail = 0;

	FILE *fp = fopen( fileName, "r" );

	if ( fp == NULL )
		return 0;

	memset( buf, 0, sizeof(buf) );
	while( !feof( fp ) )
	{
		fread( &buf[pos], 1, 1, fp );

		if ( buf[pos] == INTELHEX_START_SIGN && numRecord )
		{
			if ( !CheckCrc( buf ) )
			{
				fail = 1;
				break;
			}

			memset( buf, 0, sizeof(buf) );

			pos = 0;
			numRecord++; // Naechster Datensatz in Vorbereitung
			buf[0] = INTELHEX_START_SIGN;
		} 

		if ( pos >= sizeof(buf) )
		{
			fail = 1;
			break;
		}
		// Erster Datensatz angefangen
		if ( !numRecord )
			numRecord++;

		pos++;
	}

	fclose( fp );

	if ( !fail )
	{
		if ( !CheckCrc( buf ) )
		{
			fail = 1;
		}
		else
		{
			// Abschlusszeile pruefen (End-of-file record) 
			if ( GetType( buf ) != TypeEOF )
				fail = 1;
		}
	}

	return !fail;
}

unsigned char FileIntelHex::GetCount( const char *data )
{
	return HexToUINT8( &data[INTELHEX_COUNT] );
}

unsigned short FileIntelHex::GetAddress( const char *data )
{
	return HexToUINT16( &data[INTELHEX_ADDRESS] );
}

unsigned char FileIntelHex::GetType( const char *data )
{
	return HexToUINT8( &data[INTELHEX_TYPE] );
}

unsigned char FileIntelHex::GetData( const char *data, unsigned char *dest )
{// vom Datensatztyp unabhaengig
	unsigned short num, length = GetCount( data );
	if ( length > 0xFF )
		return 0;

	for ( num = 0; num < length; ++num )
		dest[num] = HexToUINT8( &data[INTELHEX_DATA + 2 * num] );

	return (unsigned char)num;
}

unsigned char FileIntelHex::GetData8( const char *data, unsigned char num )
{
	if ( num > GetCount( data ) )
		return 0;

	return HexToUINT8( &data[INTELHEX_DATA + 2 * num] );
}

unsigned short FileIntelHex::GetData16( const char *data, unsigned char num )
{
	if ( (num + 1) > GetCount( data ) )  
		return 0;

	return HexToUINT16( &data[INTELHEX_DATA + 2 * num] );
}

std::string FileIntelHex::errorString( EnumErrorState state )
{
	std::string text;

	switch ( state )
	{
		case None:                  text = "Kein Fehler im Zustand."; break;
		case FileOpen:              text = "Oeffnen der Datei gescheitert."; break;
		case RecordEmpty:           text = "Die uebergebene Datenzeile ist leer."; break;
		case RecordCrc:             text = "Der Berechnete CRC stimmt nicht mit dem im Datensatz ueberein."; break;
		case RecordFormat:          text = "Der Datensatz weist fehler auf. Puffer zu kurz, lang oder fehlerhafte Laengenabgabe."; break;
		case RecordBufferOverflow:  text = "Der Datensatz geht ueber die Grenze des uebergebenen Puffers hinweg."; break;
		case FileRead:              text = "Es ist ein Dateifehler beim Lesen der Daten aufgetreten."; break;
		case FileEmpty:             text = "In der Datei liegt kein Datensatz vor."; break;
		case CrcFail:               text = "CRC von Daten falsch."; break;
		case Format:                text = "Bereichsfehler bei eingelesenen Daten usw.."; break;
		case NoData:                text = "Keine Daten beim laden einer Datei."; break;
		case LoadDisplayData:       text = "Laden der Anzeigedaten gescheitert."; break;
		case LoadDeviceText:        text = "Laden der Anzeigetexte gescheitert."; break;
		case DeviceUnknown:         text = "Unbekannter Geraetetyp der Geraetedatei."; break;
		case ProgramOverflow:       text = "Programmspeicher Ueberlauf."; break;
		case RecordMake:            text = "Erzeugen eines IntelHex Datensatzes ist fehlgeschlagen."; break;
		case CryptFixFail:          text = "Verschluesselungswert ungueltig."; break;
		case CryptNone:             text = "Entschluesselungsversuch ohne vorherige Verschluesselung."; break;
		case CryptIs:               text = "Verschluesselungsversuch obwohl bereits verschluesselt."; break;
		case IniKey:                text = "Schluesselwert aus Inidatei konnte nicht geladen werden."; break;
		case IniKeyValue:           text = "Schluesselwert aus Inidatei liegt ausserhalb des gueltigen Berreiches."; break;
		case Size:                  text = "Groesse der moeglichen Datenmenge ueberschritten."; break;
		case Version:               text = "Version der Geraetedatei ist aktueller als erwartet."; break;
		case RecordEndMake:         text = "Daten oder Endzeile fuer Hexdatei konnte nicht erstellt werden."; break;
		case HeadCrc:               text = "Der CRC fuer den Dateikopf schlug fehl."; break;
		case FirmwareVersionLength: text = "Zeichenkette fuer die Firmwareversion ist zu kurz (< 8) oder zu lang (> 32)."; break;
		case ProgramCrc:            text = "Der CRC fuer den Programm-Flash schlug fehl."; break;
		case FunctionUnavailable:   text = "Funktion wird nicht unterstuetzt."; break;
		case LoadRawFirmware:       text = "Laden der Rohen Firmwaredaten gescheitert."; break;
		case VersionFile:           text = "Keine gueltige Dateiversionsangabe vorhanden."; break;
		case AsText:                text = "Fehlertext steht ueber string zur Verfuegung."; break;
		case FileFormat:            text = "Das Dateiformat ist nicht wie erwartet Intel-Hex."; break;
		case LineFormat:            text = "Das Format einer Datenzeile ist falsch."; break;
		case AllocMem:              text = "Speicherallokation konnte nicht durchgefuehrt werden."; break;
		case NoFileMatch:           text = "Es wurde keine zutreffende Geraetedatei in diesem Archiv gefunden."; break;
		case SerializeFileData:     text = "Verarbeitung der Dateidaten ist gescheitert."; break;
		default:                    text = "Unbekannter Fehlerzustand."; break;
	}

	return boost::str( boost::format( "(%1%): %2%" ) % state % text );
}

